# ENH in the article, most schemes are given only in single precision. quadpy adds
# symbolic expressions

import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import T2Scheme, register

source = article(
    authors=["P. Hillion"],
    title="Numerical Integration on a Triangle",
    journal="International Journal for Numerical Methods in Engineering",
    volume="11",
    pages="797-815",
    year="1977",
    url="https://doi.org/10.1002/nme.1620110504",
)


def hillion_01():
    d = {"centroid": [[1]]}
    return T2Scheme("Hillion 1", d, 1, source)


def hillion_02():
    d = {"d3_aa": [[frac(1, 3)], [frac(1, 2)]]}
    return T2Scheme("Hillion 2", d, 2, source)


def hillion_03():
    d = {"d3_aa": [[frac(1, 3)], [frac(1, 6)]]}
    return T2Scheme("Hillion 3", d, 2, source)


def hillion_04():
    a0, a1 = ((3 + i * sqrt(3)) / 8 for i in [+1, -1])
    d = {"s2_static": [[frac(1, 9)], [0]], "swap_ab": [[frac(4, 9)], [a0], [a1]]}
    return T2Scheme("Hillion 4", d, 2, source)


def hillion_05():
    a0, a1 = ((3 + i * sqrt(3)) / 8 for i in [+1, -1])
    d = {
        "s2_static": [[frac(1, 9)], [frac(2, 3)]],
        "swap_ab": [[frac(4, 9)], [frac(2, 3) - a0], [frac(2, 3) - a1]],
    }
    return T2Scheme("Hillion 5", d, 2, source)


def hillion_06():
    lm, mu = ((2 + i * sqrt(2 + i * sqrt(3))) / 6 for i in [+1, -1])
    d = {
        "swap_ab": [
            [frac(1, 4), frac(1, 4)],
            [lm, frac(2, 3) - lm],
            [mu, frac(2, 3) - mu],
        ]
    }
    return T2Scheme("Hillion 6", d, 2, source)


def hillion_07():
    pm = np.array([+1, -1])

    a, b = (6 + sqrt(2) + pm * sqrt(6 * (3 + 2 * sqrt(2)))) / 20
    c, d = (6 - sqrt(2) + pm * sqrt(6 * (3 - 2 * sqrt(2)))) / 20
    w1 = (2 - 3 * (b + c)) / 6 / (a + d - b - c)
    w2 = (2 - 3 * (a + d)) / 6 / (b + c - a - d)

    d = {"swap_ab": [[w1, w2], [a, c], [d, b]]}
    return T2Scheme("Hillion 7", d, 3, source, 2.220e-16)


def hillion_08():
    lambda2, lambda3 = ((32 + i * 2 * sqrt(46)) / 105 for i in [+1, -1])
    w1, w2 = (2 * (3266 + i * 19 * sqrt(46)) / 17664 for i in [+1, -1])
    d = {
        "swap_ab": [[2 * frac(25, 384)], [0], [frac(4, 5)]],
        "s2_static": [[w1, w2], [lambda2, lambda3]],
    }
    return T2Scheme("Hillion 8", d, 3, source)


def hillion_09():
    # ERR the article is missing the minus sign
    d = {"centroid": [[-frac(9, 16)]], "d3_aa": [[frac(25, 48)], [frac(1, 5)]]}
    return T2Scheme("Hillion 9", d, 3, source)


def hillion_10():
    lambda1, lambda2 = ((16 + i * 2 * sqrt(14)) / 25 for i in [+1, -1])
    w1, w2 = (2 * (161 + i * 17 * sqrt(14)) / 2688 for i in [+1, -1])
    d = {
        "swap_ab": [[w2, w1], [lambda1, 0], [0, lambda2]],
        "s2_static": [[frac(25, 48)], [frac(2, 5)]],
    }
    return T2Scheme("Hillion 10", d, 3, source)


register(
    [
        hillion_01,
        hillion_02,
        hillion_03,
        hillion_04,
        hillion_05,
        hillion_06,
        hillion_07,
        hillion_08,
        hillion_09,
        hillion_10,
    ]
)



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["Gatermann, Karin"],
    title="The construction of symmetric cubature formulas for the square and the triangle",
    journal="Computing",
    month="sep",
    year="1988",
    volume="40",
    number="3",
    pages="229–240",
    url="https://doi.org/10.1007/BF02251251",
)

this_dir = pathlib.Path(__file__).resolve().parent


def gatermann():
    return _read(this_dir / "gatermann.json", source)


register([gatermann])



from ._cools_haegemans import cools_haegemans_1, cools_haegemans_2, cools_haegemans_3
from ._lu_darmofal import (
    lu_darmofal_1,
    lu_darmofal_2,
    lu_darmofal_3,
    lu_darmofal_4a,
    lu_darmofal_4b,
)
from ._mcnamee_stenger import (
    mcnamee_stenger_3,
    mcnamee_stenger_5,
    mcnamee_stenger_7a,
    mcnamee_stenger_7b,
    mcnamee_stenger_9a,
    mcnamee_stenger_9b,
)
from ._phillips import phillips
from ._stenger import (
    stenger_7a,
    stenger_7b,
    stenger_9a,
    stenger_9b,
    stenger_11a,
    stenger_11b,
)
from ._stroud import (
    stroud_enr2_3_1,
    stroud_enr2_3_2,
    stroud_enr2_5_1a,
    stroud_enr2_5_1b,
    stroud_enr2_5_2,
    stroud_enr2_5_3,
    stroud_enr2_5_4,
    stroud_enr2_5_5a,
    stroud_enr2_5_5b,
    stroud_enr2_5_6,
    stroud_enr2_7_1a,
    stroud_enr2_7_1b,
    stroud_enr2_7_2,
    stroud_enr2_7_3a,
    stroud_enr2_7_3b,
    stroud_enr2_9_1a,
    stroud_enr2_9_1b,
    stroud_enr2_11_1a,
    stroud_enr2_11_1b,
)
from ._stroud_1967_5 import stroud_1967_5_a, stroud_1967_5_b
from ._stroud_1967_7 import stroud_1967_7_2a, stroud_1967_7_2b, stroud_1967_7_4
from ._stroud_secrest import (
    stroud_secrest_1,
    stroud_secrest_2,
    stroud_secrest_3,
    stroud_secrest_4,
)
from ._xiu import xiu

__all__ = [
    "cools_haegemans_1",
    "cools_haegemans_2",
    "cools_haegemans_3",
    "lu_darmofal_1",
    "lu_darmofal_2",
    "lu_darmofal_3",
    "lu_darmofal_4a",
    "lu_darmofal_4b",
    "mcnamee_stenger_3",
    "mcnamee_stenger_5",
    "mcnamee_stenger_7a",
    "mcnamee_stenger_7b",
    "mcnamee_stenger_9a",
    "mcnamee_stenger_9b",
    "phillips",
    "stenger_7a",
    "stenger_7b",
    "stenger_9a",
    "stenger_9b",
    "stenger_11a",
    "stenger_11b",
    "stroud_enr2_3_1",
    "stroud_enr2_3_2",
    "stroud_enr2_5_1a",
    "stroud_enr2_5_1b",
    "stroud_enr2_5_2",
    "stroud_enr2_5_3",
    "stroud_enr2_5_4",
    "stroud_enr2_5_5a",
    "stroud_enr2_5_5b",
    "stroud_enr2_5_6",
    "stroud_enr2_7_1a",
    "stroud_enr2_7_1b",
    "stroud_enr2_7_2",
    "stroud_enr2_7_3a",
    "stroud_enr2_7_3b",
    "stroud_enr2_9_1a",
    "stroud_enr2_9_1b",
    "stroud_enr2_11_1a",
    "stroud_enr2_11_1b",
    "stroud_1967_5_a",
    "stroud_1967_5_b",
    "stroud_1967_7_2a",
    "stroud_1967_7_2b",
    "stroud_1967_7_4",
    "stroud_secrest_1",
    "stroud_secrest_2",
    "stroud_secrest_3",
    "stroud_secrest_4",
    "xiu",
]



import numpy as np
import sympy

from ..helpers import article
from ._albrecht_collatz import albrecht_collatz as hammer_stroud_11_2
from ._helpers import S2Scheme, register
from ._peirce_1956 import peirce_1956_1, peirce_1956_3
from ._radon import radon

_source = article(
    authors=["Preston C. Hammer", "Arthur H. Stroud"],
    title="Numerical Evaluation of Multiple Integrals II",
    journal="Math. Comp.",
    volume="12",
    number="64",
    year="1958",
    month="oct",
    pages="272-280",
    url="https://doi.org/10.1090/S0025-5718-1958-0102176-6",
)

frac = sympy.Rational
pi = sympy.pi
sqrt = np.vectorize(sympy.sqrt)
pm_ = np.array([+1, -1])
cos = np.vectorize(sympy.cos)
sin = np.vectorize(sympy.sin)


def hammer_stroud_12_2():
    d = {
        "zero2": [[frac(1, 6)]],
        "d4_a0": [[frac(1, 6)], [sqrt(frac(1, 2))]],
        "sxy": [[frac(1, 24)], [sqrt(frac(1, 2))], [sqrt(frac(1, 2))]],
    }
    return S2Scheme("Hammer-Stroud 12-2", d, 5, _source)


def hammer_stroud_13_2():
    return peirce_1956_1()


def hammer_stroud_17():
    # DUP ENH This is Radon's formula.
    return radon(0)


def hammer_stroud_18():
    # ENH The article only gives floats, but really this is the spherical-product gauss
    # formula as described in Strouds book, S2 7-2.
    #
    # data = [
    #     (frac(1, 16), fs([0.4247082002778669, 0.1759198966061612])),
    #     (frac(1, 16), fs([0.8204732385702833, 0.3398511429799874])),
    # ]
    r1, r2 = sqrt((3 - pm_ * sqrt(3)) / 6)

    d = {"d8.0": [[frac(1, 16), frac(1, 16)], [r1, r2]]}
    return S2Scheme("Hammer-Stroud 18", d, 7, _source)


def hammer_stroud_19():
    sqrt6 = sqrt(6)
    alpha1 = (16 + sqrt6) / 288
    alpha2 = (137 - 32 * sqrt6) / 1818
    alpha3 = (520 + 155 * sqrt6) / 3636 / 8

    a = sqrt((6 + sqrt6) / 10)

    d = {
        "zero2": [[frac(1, 9)]],
        "d4_ab": [
            [alpha1, alpha3],
            [0.5505043204538557, 0.7932084745126058],
            [0.2280263556769715, 0.4645097310495256],
        ],
        "d4_a0": [[alpha2], [a]],
    }
    return S2Scheme("Hammer-Stroud 19", d, 9, _source)


def hammer_stroud_20():
    # ENH Also Peirce's formula, even given symbolically.
    return peirce_1956_3()


def hammer_stroud_21():
    alpha0 = 0.0341505695624825 / np.pi
    alpha1 = 0.0640242008621985 / np.pi
    alpha2 = 0.0341505695624825 / np.pi

    d = {
        "d4_ab": [
            [alpha0, alpha1, alpha1, alpha1, alpha1, alpha2, alpha2, alpha2],
            [
                0.2584361661674054,
                0.5634263397544869,
                0.4776497869993547,
                0.8028016728473508,
                0.6805823955716280,
                0.2190916025980981,
                0.9461239423417719,
                0.8020851487551318,
            ],
            [
                0.0514061496288813,
                0.1120724670846205,
                0.3191553840796721,
                0.1596871812824163,
                0.4547506180649039,
                0.1463923286035535,
                0.1881957532057769,
                0.5359361621905023,
            ],
        ],
    }
    return S2Scheme("Hammer-Stroud 21", d, 15, _source)


register(
    [
        hammer_stroud_11_2,
        hammer_stroud_12_2,
        hammer_stroud_13_2,
        hammer_stroud_17,
        hammer_stroud_18,
        hammer_stroud_19,
        hammer_stroud_20,
        hammer_stroud_21,
    ]
)



import math

import sympy


def volume(n, r=1, symbolic=False):
    pi = sympy.pi if symbolic else math.pi

    def _recurrence(n):
        assert n >= 1
        if n == 1:
            return 2
        elif n == 2:
            return 2 * pi * r
        return _recurrence(n - 2) * 2 * pi / (n - 2) * r**2

    return _recurrence(n)


def integrate_monomial(exponents, symbolic=False):
    exponents = list(exponents)
    n = len(exponents)
    n = sympy.S(n) if symbolic else n

    if any(k % 2 == 1 for k in exponents):
        return 0

    def _recurrence(exponents):
        if all(k == 0 for k in exponents):
            return volume(n, symbolic=symbolic)

        # find first nonzero
        idx, k0 = next((i, k) for i, k in enumerate(exponents) if k > 0)
        k2 = exponents.copy()
        k2[idx] -= 2
        return _recurrence(k2) * (k0 - 1) / (sum(exponents) + n - 2)

    return _recurrence(exponents)
